# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Credential-related classes.

    CredentialByteArray: Uses byte_array so cred can be zero'ed out afterwards as
    a form of cleanup. Any such efforts at cleanup are *not* security for several
    good reasons, one of which is that CredentialByteArray is not the only party
    handling secrets, where some others such as the cryptography/keyring packages
    use bytes immutable so do not allow, from Python perspective, the zero'ing
    out the memory area they populate with keys/passwords.

    A best effort toward minimizing the time frame when secrets are exposed would be
    to have all parties handling secrets to use a "SecureByteArray" of sorts, which
    would allow memory locking, enforce clearing at cleanup, etc. Stil, None of
    that would remove the need to keep the machine/process secure, but would merely
    minimize the issue caused by immutables which themselves may well leave secrets
    lying around for much longer time periods, if not spreading copies around beyond
    some original.

    Therefore, to achieve security when using a Python process that handles secrets
    in the clear at *any* time for *however* long a duration, the machine and process
    space must be kept clear of threats.
"""

from abc import ABC, abstractmethod
import base64
import os
from typing import Callable, Union
from hashlib import pbkdf2_hmac
import keyring
import pwinput

from atbu.common.util_helpers import is_valid_base64_string
from atbu.common.aes_cbc import (
    AES_CBC_Base,
    AesCbcPaddingDecryptor,
    AesCbcPaddingEncryptor,
)

from .constants import *
from .exception import (
    AtbuException,
    CredentialNotFoundError,
    InvalidBase64StringError,
    PasswordAuthenticationFailure,
    CredentialInvalid,
    CredentialSecretDerivationError,
)
from .yubikey_helpers import (
    get_max_challenge_size,
    is_yubikey_required,
    is_a_yubikey_present,
    challenge_response,
    YubiKeyNotPressedTimeout,
)

_MAX_PASSWORD = 100


def zero_bytearray(ba: bytearray):
    for v in [0xCC, 0x55, 0x00]:
        for i, _ in enumerate(ba):
            ba[i] = v


class CredentialByteArray(bytearray):
    """A bytearray with override to allow zeroing out of bytearray
    on demand or when deleted.
    """

    def __del__(self):
        self.zero_array()
        if hasattr(super(), "__del__"):
            super().__del__()

    def zero_array(self):
        zero_bytearray(self)

    def get_portion(self, start_pos=None, end_pos=None):
        if not start_pos:
            start_pos = 0
        if not end_pos:
            end_pos = len(self)
        portion_len = end_pos - start_pos
        r = CredentialByteArray(portion_len)
        idx_r = 0
        for i in range(start_pos, end_pos):
            r[idx_r] = self[i]
            idx_r += 1
        return r

    def split(self, sep, maxsplit) -> list:
        result = []
        for ba in super().split(sep=sep, maxsplit=maxsplit):
            result.append(CredentialByteArray(ba))
            zero_bytearray(ba)
        return result

    def to_serialization_dict(self) -> dict:
        return self.hex()

    def from_serialization_dict(self, hex_str):
        self.fromhex(hex_str)

    @staticmethod
    def create_from_string(the_string: str):
        return CredentialByteArray(the_string.encode("utf-8"))


class Credential:
    """A Credential instance is one credential, which can be an encryption key or
    any other kind of secret that can be represented as bytes.

    The secret, represented by Credential.the_key, can itself be password-protected
    using key derived by PBKDF2.
    """

    attr_name_list = [
        "password",
        "salt",
        "pbkdf2_result",
        "password_auth_hash",
        "key_encryption_key",
        "iv",
        "encrypted_key",
        "the_key",
    ]

    def __init__(
        self,
        password: Union[str, CredentialByteArray] = None,
        salt: CredentialByteArray = None,
        password_auth_hash: CredentialByteArray = None,
        key_encryption_key: CredentialByteArray = None,
        iv: CredentialByteArray = None,
        encrypted_key: CredentialByteArray = None,
        the_key: CredentialByteArray = None,
        **kwargs,
    ) -> None:

        super().__init__(**kwargs)

        self.password = None  # The password used by PBKDF2 to derive the KEK and password auth hash.
        self.salt = None  # The salt used by PBKDF2.
        self.password_auth_hash = None  # The password auth hash resulting from PBKDF2.
        self.key_encryption_key = None  # The KEK resulting from PBKDF2.
        self.iv = None  # The IV associated with encrypting/decrypting the key.
        self.encrypted_key = None  # The encrypted (protected) key.
        self.the_key = None  # The unencrypted (unprotected) key (aka "secret").
        self.pbkdf2_result = None  # The raw PBKDF2 result.
        self.work_factor = None  # The work factor used by PBKDF2.

        self.clear()
        self.set(
            password=password,
            salt=salt,
            password_auth_hash=password_auth_hash,
            key_encryption_key=key_encryption_key,
            iv=iv,
            encrypted_key=encrypted_key,
            the_key=the_key,
        )

    @property
    def is_private_key_possible(self):
        """True if a private key is achievable either as is or with a password,
        False if not possible no matter what. See is_password_required.
        Generally, use is_private_key_possible to see if a private is available
        at all by any means. Then, if this is True, use is_password_required to
        determine is the user should be prompted for a password.
        """
        if self.the_key is not None:
            return True
        if self.encrypted_key is None:
            return False
        if self.salt is None or self.iv is None:
            return False
        return True

    @property
    def is_password_required(self):
        """True if password is required, False if not required.
        See is_private_key_possible.
        """
        if self.the_key is not None:
            return False
        if self.key_encryption_key is not None:
            return False
        if self.password is not None:
            return False
        return True

    @property
    def is_password_protected(self):
        """True if the credential is protected by a password, False if not
        protected by a password. This property indications protection regardless
        of whether or not the key is currently unprotected (i.e., user already
        entered password, key already decrypted). This property therefore reflects
        generally whether or not the crednetial is password protected.
        """
        return self.password is not None or self.encrypted_key is not None

    @property
    def is_private_key_ready(self):
        return self.the_key is not None

    def __del__(self):
        self.clear()
        if hasattr(super(), "__del__"):
            super().__del__()

    def clear(self) -> None:
        for member in Credential.attr_name_list:
            if hasattr(self, member) and isinstance(member, CredentialByteArray):
                cba: CredentialByteArray = getattr(self, member)
                cba.zero_array()
        self.password = None
        self.salt = None
        self.password_auth_hash = None
        self.key_encryption_key = None
        self.iv = None
        self.encrypted_key = None
        self.the_key = None
        self.pbkdf2_result = None
        self.work_factor = PBKDF2_WORK_FACTOR
        self.clear_password_protection()

    def clear_password(self):
        if self.password:
            self.password.zero_array()
        self.password = None

    def set(
        self,
        password: Union[str, CredentialByteArray] = None,
        salt: CredentialByteArray = None,
        password_auth_hash: CredentialByteArray = None,
        key_encryption_key: CredentialByteArray = None,
        iv: CredentialByteArray = None,
        encrypted_key: CredentialByteArray = None,
        the_key: CredentialByteArray = None,
    ):
        if isinstance(password, str):
            password = CredentialByteArray(password.encode())
        if password is not None:
            if not isinstance(password, CredentialByteArray):
                raise ValueError("The password is not a CredentialByteArray.")
            if self.iv is None or self.salt is None:
                self.prepare_for_new_password()
            self.password = password

        if salt is not None:
            if not isinstance(salt, CredentialByteArray):
                raise ValueError("The salt is not a CredentialByteArray.")
            self.salt = salt

        if password_auth_hash is not None:
            if not isinstance(password_auth_hash, CredentialByteArray):
                raise ValueError("The password_auth_hash is not a CredentialByteArray.")
            self.password_auth_hash = password_auth_hash

        if key_encryption_key is not None:
            if not isinstance(key_encryption_key, CredentialByteArray):
                raise ValueError("The key_encryption_key is not a CredentialByteArray.")
            self.key_encryption_key = password_auth_hash

        if iv is not None:
            if not isinstance(iv, CredentialByteArray):
                raise ValueError("The iv is not a CredentialByteArray.")
            self.iv = iv

        if encrypted_key is not None:
            if not isinstance(encrypted_key, CredentialByteArray):
                raise ValueError("The encrypted_key is not a CredentialByteArray.")
            self.encrypted_key = encrypted_key

        if the_key is not None:
            if not isinstance(the_key, CredentialByteArray):
                raise ValueError("The key is not a CredentialByteArray.")
            self.the_key = the_key

    def prepare_for_new_password(self):
        self.clear_password_protection()
        self.salt = CredentialByteArray(os.urandom(DEFAULT_AES_KEY_BIT_LENGTH // 8))
        self.iv = CredentialByteArray(os.urandom(AES_CBC_Base.BLOCK_SIZE))

    def clear_password_protection(self):
        if self.salt:
            self.salt.zero_array()
        if self.iv:
            self.iv.zero_array()
        if self.password:
            self.password.zero_array()
        if self.password_auth_hash:
            self.password_auth_hash.zero_array()
        if self.key_encryption_key:
            self.key_encryption_key.zero_array()
        if self.encrypted_key:
            self.encrypted_key.zero_array()
        self.salt = None
        self.iv = None
        self.password = None
        self.password_auth_hash = None
        self.key_encryption_key = None
        self.encrypted_key = None

    def encrypt_key(self):
        self.password_auth_hash = self._derive_key_encryption_key()
        enc = AesCbcPaddingEncryptor(self.key_encryption_key, self.iv)
        encrypted_key_bytes = enc.update(input_bytes=self.the_key) + enc.finalize()
        self.encrypted_key = CredentialByteArray(encrypted_key_bytes)

    def decrypt_key(self):
        if self.encrypted_key is None:
            raise ValueError("No encrypted key to decrypt.")
        derived_password_auth_hash = self._derive_key_encryption_key()
        if self.password_auth_hash != derived_password_auth_hash:
            raise PasswordAuthenticationFailure(
                "The password and/or stored secrets do not match."
            )
        dec = AesCbcPaddingDecryptor(key=self.key_encryption_key, IV=self.iv)
        decrypted_key_bytes = (
            dec.update(input_bytes=self.encrypted_key) + dec.finalize()
        )
        self.the_key = CredentialByteArray(decrypted_key_bytes)

    def _derive_key_encryption_key(self):
        if self.password is None:
            raise ValueError("A password is required to derive the key encryption key.")
        if self.salt is None:
            raise ValueError("A salt is required to derive the key encryption key.")
        if self.work_factor is None:
            raise ValueError("A work factor is needed for PBDKF2.")
        self.pbkdf2_result = CredentialByteArray(
            pbkdf2_hmac(
                hash_name="sha512",
                password=self.password,
                salt=self.salt,
                iterations=self.work_factor,
                dklen=int(512 / 8),
            )
        )

        kek_byte_size = DEFAULT_AES_KEY_BIT_LENGTH // 8
        self.key_encryption_key = self.pbkdf2_result.get_portion(0, kek_byte_size)
        password_auth_hash = self.pbkdf2_result.get_portion(
            kek_byte_size, kek_byte_size * 2
        )
        return password_auth_hash

    def get_as_bytes(
        self,
        include_work_factor: bool = False,
        include_salt: bool = False,
        include_password_auth_hash: bool = False,
        include_IV: bool = False,
        include_key_encryption_key: bool = False,
        include_key: bool = False,
        include_encrypted_key: bool = False,
    ) -> CredentialByteArray:

        #
        # In the following, we ultimately get everything into a textual/hex format
        # where fields are separated by the equal sign (=) and commas (,). After
        # TODO: Loop back around to encode with struct pack/unpack.
        #

        r = CredentialByteArray()

        if include_work_factor:
            if len(r) > 0:
                r += b","
            r += b"W="
            r += (
                int(PBKDF2_WORK_FACTOR)
                .to_bytes(length=4, byteorder="little")
                .hex()
                .encode()
            )

        if include_salt:
            if len(r) > 0:
                r += b","
            r += b"S="
            r += self.salt.hex().encode()

        if include_password_auth_hash:
            if len(r) > 0:
                r += b","
            r += b"H="
            r += self.password_auth_hash.hex().encode()

        if include_key_encryption_key:
            if len(r) > 0:
                r += b","
            r += b"E="
            r += self.key_encryption_key.hex().encode()

        if include_IV:
            if len(r) > 0:
                r += b","
            r += b"I="
            r += self.iv.hex().encode()

        if include_encrypted_key:
            if len(r) > 0:
                r += b","
            r += b"C="
            r += self.encrypted_key.hex().encode()

        if include_key:
            if len(r) > 0:
                r += b","
            r += b"K="
            r += self.the_key.hex().encode()

        if len(r) == 0:
            raise ValueError(f"Requesting no secrets.")
        return r

    def set_from_bytes(self, cred_bytes: CredentialByteArray):
        cred_dict = dict(
            kv_pair.split("=") for kv_pair in cred_bytes.decode().split(",")
        )
        for k, v in cred_dict.items():
            c: CredentialByteArray = CredentialByteArray(bytearray.fromhex(v))
            if k == "W":
                self.work_factor = int.from_bytes(bytes=c, byteorder="little")
                if self.work_factor != PBKDF2_WORK_FACTOR:
                    # Until handling for this is in place.
                    raise CredentialInvalid(
                        f"The PBDKF2 work factor "
                        f"{self.work_factor} is invalid. "
                        f"Only {PBKDF2_WORK_FACTOR} is supported."
                    )
            elif k == "S":
                self.salt = c
            elif k == "H":
                self.password_auth_hash = c
            elif k == "E":
                self.key_encryption_key = c
            elif k == "I":
                self.iv = c
            elif k == "C":
                self.encrypted_key = c
            elif k == "K":
                self.the_key = c
            else:
                raise ValueError(f"Cannot interpret value code '{k}'.")

    def get_enc_key_material_as_bytes(self) -> CredentialByteArray:
        return self.get_as_bytes(
            include_work_factor=True,
            include_salt=True,
            include_password_auth_hash=True,
            include_IV=True,
            include_encrypted_key=True,
        )

    def get_unenc_key_material_as_bytes(self) -> CredentialByteArray:
        return self.get_as_bytes(include_key=True)

    def get_material_as_bytes(self) -> CredentialByteArray:
        items_found = []
        expected_items = 4
        if self.salt is not None:
            items_found.append("salt")
        if self.iv is not None:
            items_found.append("iv")
        if self.password_auth_hash is not None:
            items_found.append("password_auth_hash")
        if self.encrypted_key is not None:
            items_found.append("encrypted_key")
        if len(items_found) != 0 and len(items_found) != expected_items:
            raise CredentialInvalid(
                f"The credential has partial password-protection settings: "
                f"count={len(items_found)} expected={expected_items} items_found={items_found}"
            )
        if len(items_found) == 0:
            return self.get_unenc_key_material_as_bytes()
        return self.get_enc_key_material_as_bytes()

    @classmethod
    def create_credential_from_bytes(cls, cred_bytes: CredentialByteArray):
        c = cls()
        c.set_from_bytes(cred_bytes=cred_bytes)
        return c


class CredentialAesKey(Credential):
    def __init__(
        self,
        password: Union[str, CredentialByteArray] = None,
        salt: CredentialByteArray = None,
        password_auth_hash: CredentialByteArray = None,
        key_encryption_key: CredentialByteArray = None,
        iv: CredentialByteArray = None,
        encrypted_key: CredentialByteArray = None,
        the_key: CredentialByteArray = None,
        key_bit_size=DEFAULT_AES_KEY_BIT_LENGTH,
        **kwargs,
    ) -> None:
        super().__init__(
            password=password,
            salt=salt,
            password_auth_hash=password_auth_hash,
            key_encryption_key=key_encryption_key,
            iv=iv,
            encrypted_key=encrypted_key,
            the_key=the_key,
            **kwargs,
        )
        self.key_bit_size = None
        self.key_byte_size = None

        self.clear(key_bit_size=key_bit_size)
        self.set(
            password=password,
            salt=salt,
            password_auth_hash=password_auth_hash,
            key_encryption_key=key_encryption_key,
            iv=iv,
            encrypted_key=encrypted_key,
            the_key=the_key,
            key_bit_size=key_bit_size,
        )

    def set(
        self,
        password: Union[str, CredentialByteArray] = None,
        salt: CredentialByteArray = None,
        password_auth_hash: CredentialByteArray = None,
        key_encryption_key: CredentialByteArray = None,
        iv: CredentialByteArray = None,
        encrypted_key: CredentialByteArray = None,
        the_key: CredentialByteArray = None,
        key_bit_size=None,
    ):
        super().set(
            password=password,
            salt=salt,
            password_auth_hash=password_auth_hash,
            key_encryption_key=key_encryption_key,
            iv=iv,
            encrypted_key=encrypted_key,
            the_key=the_key,
        )
        if key_bit_size is not None:
            self._set_key_bit_size(key_bit_size=key_bit_size)

    def clear(self, key_bit_size=DEFAULT_AES_KEY_BIT_LENGTH) -> None:
        super().clear()
        self._set_key_bit_size(key_bit_size=key_bit_size)

    def _set_key_bit_size(self, key_bit_size):
        if key_bit_size is None or key_bit_size not in ALLOWED_AES_KEY_BIT_LENGTHS:
            raise ValueError(
                f"The specified key length of '{key_bit_size} is not one of the "
                f"available choices of {ALLOWED_AES_KEY_BIT_LENGTHS}."
            )
        self.key_bit_size = key_bit_size
        self.key_byte_size = int(key_bit_size / 8)

    def create_key(
        self,
        the_key: CredentialByteArray = None,
        key_bit_size=DEFAULT_AES_KEY_BIT_LENGTH,
    ) -> None:
        self.clear(key_bit_size=key_bit_size)
        if not the_key:
            the_key = CredentialByteArray(os.urandom(self.key_byte_size))
        else:
            if (
                not isinstance(the_key, CredentialByteArray)
                or len(the_key) != self.key_byte_size
            ):
                raise ValueError(
                    f"The key must be the specified length of {self.key_bit_size} bits"
                )
        self.the_key = the_key


class DescribedCredential:
    def __init__(
        self,
        credential: Credential = None,
        config_name: str = "unknown",
        credential_name: str = "unknown",
        credential_kind: str = "unknown",
    ) -> None:
        self.credential = credential
        self.config_name = config_name
        self.credential_name = credential_name
        self.credential_kind = credential_kind

    def get_credential_base64(self) -> CredentialByteArray:
        return CredentialByteArray(
            base64.b64encode(self.credential.get_material_as_bytes())
        )

    @staticmethod
    def create_from_base64(
        config_name: str,
        credential_name: str,
        password_type_code: str,
        cba_password_base64: CredentialByteArray,
    ):
        if len(password_type_code) == 0:
            raise CredentialInvalid(
                f"The stored credential is not in the expected format. "
                f"The password type code is not present. "
                f"config_name={config_name} credential_name={credential_name}"
            )

        if len(cba_password_base64) == 0:
            raise CredentialInvalid(
                f"The stored credential is not in the expected format. "
                f"The password is not present. "
                f"config_name={config_name} credential_name={credential_name}"
            )

        if not is_valid_base64_string(cba_password_base64):
            raise InvalidBase64StringError(
                f"Expected stored credential to be in base64 format."
            )

        if password_type_code[0] not in PASSWORD_KIND_CHAR_TO_KIND:
            raise CredentialInvalid(
                f"The stored credential is not in the expected format. "
                f"The password type code '{password_type_code}' is unknown. "
                f"config_name={config_name} credential_name={credential_name}"
            )

        password_type = PASSWORD_KIND_CHAR_TO_KIND[password_type_code[0]]

        cba_password = CredentialByteArray(base64.b64decode(cba_password_base64))

        if credential_name == CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION:
            if password_type != CONFIG_PASSWORD_KIND_ACTUAL:
                raise CredentialInvalid(
                    f"The stored credential is type '{password_type}' when "
                    f"'{CONFIG_PASSWORD_KIND_ACTUAL}' was expected."
                )
            credential = CredentialAesKey()
            credential.set_from_bytes(cred_bytes=cba_password)
        elif credential_name == CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD:
            credential = Credential()
            credential.set_from_bytes(cred_bytes=cba_password)

        return DescribedCredential(
            credential=credential,
            config_name=config_name,
            credential_name=credential_name,
            credential_kind=password_type,
        )


def raw_cred_bytes_to_type_base64_cred_bytes(
    cred_ascii_bytes: CredentialByteArray,
) -> tuple[str, CredentialByteArray]:
    """Get the password_type_code and raw credential from complete raw persisted
    credential bytes. This can be further refined, is carried over to limit
    refactor churn.
    """
    if not isinstance(cred_ascii_bytes, CredentialByteArray):
        raise CredentialNotFoundError(f"The credential was not found: ")
    password_parts = cred_ascii_bytes.split(b":", maxsplit=1)
    if len(password_parts) != 2:
        raise CredentialInvalid(
            f"The stored credential is not in the expected two-part format."
        )
    password_type_code = str(password_parts[0], "utf-8")
    cba_password_base64 = password_parts[1]
    if password_type_code not in PASSWORD_KIND_CHAR_TO_KIND:
        raise CredentialInvalid(
            f"The credential is not in the expected format. "
            f"The password type code '{password_type_code}' is unknown. "
        )
    password_type = PASSWORD_KIND_CHAR_TO_KIND[password_type_code]
    return password_type, cba_password_base64


class CredentialStoreProvider(ABC):
    @abstractmethod
    def set_cred_bytes(
        self,
        config_name: str,
        credential_name: str,
        cred_ascii_bytes: CredentialByteArray,
    ) -> None:
        pass

    @abstractmethod
    def get_cred_bytes(
        self, config_name: str, credential_name: str
    ) -> CredentialByteArray:
        pass

    @abstractmethod
    def delete_cred_bytes(self, config_name: str, credential_name: str) -> None:
        pass


class CredentialStoreKeyringProvider(CredentialStoreProvider):
    def set_cred_bytes(
        self,
        config_name: str,
        credential_name: str,
        cred_ascii_bytes: CredentialByteArray,
    ) -> None:
        keyring.set_password(
            service_name=config_name,
            username=credential_name,
            password=str(cred_ascii_bytes, "utf-8"),
        )

    def get_cred_bytes(
        self, config_name: str, credential_name: str
    ) -> CredentialByteArray:
        password_str = keyring.get_password(
            service_name=config_name,
            username=credential_name,
        )
        if password_str is None:
            raise CredentialNotFoundError(
                f"The credential was not found: "
                f"config_name={config_name} credential_name={credential_name}"
            )
        return CredentialByteArray(password_str.encode("utf-8"))

    def delete_cred_bytes(self, config_name: str, credential_name: str) -> None:
        keyring.delete_password(
            service_name=config_name,
            username=credential_name,
        )


class ProviderCannotHandleException(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class CredentialStoreProviderList(CredentialStoreProvider):

    def __init__(self) -> None:
        super().__init__()
        self.factory_list: list[Callable] = []

    def get_factory_list(self) -> list[Callable]:
        return self.factory_list

    def append(self, provider_factory: Callable):
        self.factory_list.append(provider_factory)

    def append_default_factories(self) -> None:
        self.append(provider_factory=CredentialStoreKeyringProvider)

    def set_cred_bytes(
        self,
        config_name: str,
        credential_name: str,
        cred_ascii_bytes: CredentialByteArray,
    ) -> None:
        for provider_factory in self.factory_list:
            try:
                provider: CredentialStoreProvider = provider_factory()
                provider.set_cred_bytes(
                    config_name=config_name,
                    credential_name=credential_name,
                    cred_ascii_bytes=cred_ascii_bytes,
                )
                return
            except ProviderCannotHandleException:
                # A non-error indicating the provider cannot handle the request
                # for normal reasons (i.e., it does not represent the storage
                # definition (aka config_name) requested). All other exceptions
                # are considered abnormal.
                # Try the next provider.
                pass

    def get_cred_bytes(
        self, config_name: str, credential_name: str
    ) -> CredentialByteArray:
        for provider_factory in self.factory_list:
            try:
                provider: CredentialStoreProvider = provider_factory()
                return provider.get_cred_bytes(
                    config_name=config_name,
                    credential_name=credential_name,
                )
            except ProviderCannotHandleException:
                # A non-error indicating the provider cannot handle the request
                # for normal reasons (i.e., it does not represent the storage
                # definition (aka config_name) requested). All other exceptions
                # are considered abnormal.
                # Try the next provider.
                pass

    def delete_cred_bytes(self, config_name: str, credential_name: str):
        for provider_factory in self.factory_list:
            try:
                provider: CredentialStoreProvider = provider_factory()
                provider.delete_cred_bytes(
                    config_name=config_name,
                    credential_name=credential_name,
                )
                return
            except ProviderCannotHandleException:
                # A non-error indicating the provider cannot handle the request
                # for normal reasons (i.e., it does not represent the storage
                # definition (aka config_name) requested). All other exceptions
                # are considered abnormal.
                # Try the next provider.
                pass


_credential_provider_cls: CredentialStoreProvider = CredentialStoreKeyringProvider


def set_credential_store(store_cls: Callable):
    global _credential_provider_cls
    _credential_provider_cls = store_cls


class CredentialStore:
    def __init__(self, provider: CredentialStoreProvider = None) -> None:
        super().__init__()
        if provider is None:
            provider = _credential_provider_cls()
        self.provider = provider

    def set_credential(self, desc_cred: DescribedCredential) -> CredentialByteArray:
        if not desc_cred.config_name:
            raise CredentialInvalid(
                f"Cannot store credential. The config_name must be a non-empty string."
            )

        if not desc_cred.credential_name:
            raise CredentialInvalid(
                f"Cannot store credential. The credential_name must be a non-empty string."
            )

        if not desc_cred.credential_kind:
            raise CredentialInvalid(
                f"Cannot store credential. The credential_type must be a non-empty string."
            )

        if desc_cred.credential_kind not in PASSWORD_KINDS:
            raise CredentialInvalid(
                f"Cannot store credential. "
                f"The supplied password type '{desc_cred.credential_kind}' is invalid."
            )

        cba_password_base64 = desc_cred.get_credential_base64()
        cred_ascii_bytes = CredentialByteArray(
            f"{desc_cred.credential_kind[0]}:{str(cba_password_base64, 'utf-8')}".encode(
                "utf-8"
            )
        )

        # For any desired caller rollback, get the existing credential.
        # This assumes no format and does not propagate failre to get any
        # older credential beyond returning None.
        cba_old: CredentialByteArray = None
        try:
            cba_old = self.provider.get_cred_bytes(
                config_name=desc_cred.config_name,
                credential_name=desc_cred.credential_name,
            )
        except Exception:
            cba_old = None

        self.provider.set_cred_bytes(
            config_name=desc_cred.config_name,
            credential_name=desc_cred.credential_name,
            cred_ascii_bytes=cred_ascii_bytes,
        )

        return cba_old

    def get_credential(
        self, config_name: str, credential_name: str
    ) -> DescribedCredential:

        cred_ascii_bytes = self.provider.get_cred_bytes(
            config_name=config_name,
            credential_name=credential_name,
        )

        (
            password_type,
            cba_password_base64,
        ) = raw_cred_bytes_to_type_base64_cred_bytes(
            cred_ascii_bytes=cred_ascii_bytes,
        )

        desc_cred = DescribedCredential.create_from_base64(
            config_name=config_name,
            credential_name=credential_name,
            password_type_code=password_type,
            cba_password_base64=cba_password_base64,
        )

        return desc_cred

    def delete_credential(self, config_name: str, credential_name: str) -> None:
        self.provider.delete_cred_bytes(
            config_name=config_name,
            credential_name=credential_name,
        )


def get_base64_password_from_keyring(
    service_name: str,
    username: str,
) -> tuple[str, str, CredentialByteArray]:
    raw_password = keyring.get_password(service_name=service_name, username=username)
    if raw_password is None:
        raise CredentialNotFoundError(
            f"The credential was not found in the keyring: "
            f"service_name={service_name} username={username}"
        )
    password_parts = raw_password.split(":", maxsplit=1)
    if len(password_parts) != 2:
        raise CredentialInvalid(
            f"The stored credential is not in the expected two-part format."
        )
    password_type_code = password_parts[0]
    cba_password_base64 = CredentialByteArray(password_parts[1].encode("utf-8"))

    if len(password_type_code) == 0:
        raise CredentialInvalid(
            f"The stored credential is not in the expected format. "
            f"The password type code is not present. "
            f"service_name={service_name} username={username}"
        )

    if len(cba_password_base64) == 0:
        raise CredentialInvalid(
            f"The stored credential is not in the expected format. "
            f"The password is not present. "
            f"service_name={service_name} username={username}"
        )

    if not is_valid_base64_string(cba_password_base64):
        raise InvalidBase64StringError(
            f"Expected stored credential to be in base64 format."
        )

    if password_type_code not in PASSWORD_KIND_CHAR_TO_KIND:
        raise CredentialInvalid(
            f"The stored credential is not in the expected format. "
            f"The password type code '{password_type_code}' is unknown. "
            f"service_name={service_name} username={username}"
        )

    password_type = PASSWORD_KIND_CHAR_TO_KIND[password_type_code]

    return (password_type_code, password_type, cba_password_base64)


def default_is_password_valid(password: CredentialByteArray):
    if len(password) > _MAX_PASSWORD:
        print(f"Your password is too long.")
        print(f"The maximum length password is {_MAX_PASSWORD} UTF-8 encoded bytes.")
        return False
    if len(password) == 0:
        print("Blank passwords are not allowed, try again.")
        return False
    return True


def prompt_for_password(
    prompt,
    prompt_again=None,
    hidden: bool = True,
    is_password_valid_func: Callable[[CredentialByteArray], bool] = None,
) -> CredentialByteArray:
    if is_password_valid_func is None:
        is_password_valid_func = default_is_password_valid
    pw_input_func = pwinput.pwinput if hidden else input
    while True:
        password_attempt1 = CredentialByteArray(pw_input_func(prompt).encode("utf-8"))
        if not is_password_valid_func(password_attempt1):
            continue
        if not prompt_again:
            return password_attempt1
        password_attempt2 = CredentialByteArray(pw_input_func(prompt_again).encode())
        if password_attempt1 == password_attempt2:
            del password_attempt2
            return password_attempt1
        print("The passwords did not match, try again.")


def check_yubikey_presence_with_message(display_presence_message: bool = False):
    if not is_a_yubikey_present():
        print(
            f"IMPORTANT: No YubiKey was detected. "
            f"Please insert your YubiKey before entering your password."
        )
        return False
    if display_presence_message:
        print(f"A YubiKey was detected.")
    return True


def is_password_valid_yubikey(password: CredentialByteArray):
    if len(password) > get_max_challenge_size():
        print(f"The password you entered is {len(password)} byte when UTF-8 encoded.")
        print(
            f"The UTF-8 encoded password cannot be larger than {get_max_challenge_size()} bytes."
        )
        print(f"Please try a shorter password.")
        check_yubikey_presence_with_message()
        return False
    return check_yubikey_presence_with_message()


def prompt_for_password_with_yubikey_opt(
    prompt,
    prompt_again=None,
    hidden: bool = True,
) -> CredentialByteArray:
    is_password_valid_func = None

    while True:
        if is_yubikey_required():
            is_password_valid_func = is_password_valid_yubikey
            check_yubikey_presence_with_message(display_presence_message=True)

        password = prompt_for_password(
            prompt=prompt,
            prompt_again=prompt_again,
            hidden=hidden,
            is_password_valid_func=is_password_valid_func,
        )

        if not is_yubikey_required():
            break

        try:
            print(f"Press your key now to allow challenge/response...")
            password = CredentialByteArray(challenge_response(challenge=password))
            break
        except YubiKeyNotPressedTimeout as ex:
            print(f"You did not press your key in time. Please try again.")
        except OSError as ex:
            if is_a_yubikey_present():
                raise
            print(f"The YubiKey is no longer present.")

    return password


def prompt_for_password_unlock_credential(
    credential: Union[Credential, CredentialAesKey],
    prompt: str = None
):
    """Unlock a CredentialAesKey instance which is to make its private key/secret
    available in the clear. If the private key is password-protected, ask
    for a password. The result of calling this function is an unlocked
    credential or an exception.

        Exceptions:
            PasswordAuthenticationFailure: Too many attempts with incorrect password.
            CredentialSecretDerivationError: Something unexpected indicates the key
            is not available despite having been decrypted (likely a program bug or
            something esoteric such as system corruption).
    """
    if prompt is None:
        prompt = "Enter the password for this backup:"
    if not credential.is_private_key_possible:
        raise CredentialInvalid(
            f"The private key is not available. "
            f"The credential is invalid or corrupt."
        )
    if not credential.is_private_key_ready:
        if not credential.is_password_required:
            credential.decrypt_key()
        else:
            attempts = 5
            while True:
                password = prompt_for_password_with_yubikey_opt(
                    prompt=prompt
                )
                credential.set(password=password)
                try:
                    attempts -= 1
                    credential.decrypt_key()
                    break
                except PasswordAuthenticationFailure:
                    if attempts > 0:
                        print(f"The password appears to be invalid, try again.")
                    else:
                        print(f"Still incorrect.")
                        raise
        if not credential.is_private_key_ready:
            raise CredentialSecretDerivationError(
                "Unexpected failure, canonot access the credential."
            )



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""ATBU Exceptions.
"""

from atbu.common.exception import *


class PersistentFileInfoError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class InvalidPersistentFileInfoError(PersistentFileInfoError):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class PersistentFileInfoVersionMismatch(PersistentFileInfoError):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class InvalidStorageDefinitionName(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class InvalidStorageDefinitionFile(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class StorageDefinitionAlreadyExists(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class StorageDefinitionNotFoundError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class StorageDefinitionNotCreatedError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class StorageDefinitionNotSpecifiedError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        super().__init__(message=message, cause=cause)


class PasswordAuthenticationFailure(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class CredentialInvalid(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class CredentialSetInvalid(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class CredentialStateInvalid(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class CredentialNotFoundError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class CredentialSecretDerivationError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class CredentialSecretFileNotFoundError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class CredentialTypeNotFoundError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class CredentialRequestInvalidError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class BackupSelectionError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class BackingFileInformationNotFound(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class RestoreFilePathAlreadyExistsError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class VerifyFilePathNotFoundError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class VerifyFailure(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class BackupException(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class BackupInformationDirectoryNotFound(BackupException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class BackupInformationError(BackupException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class BackupInformationRecoveryFailed(BackupException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class BackupInformationFileTimestampNotFound(BackupException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class BackupFileHeaderInvalid(BackupException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class PreambleParsingError(BackupException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class BackupFileInformationError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class YubiKeyBackendNotAvailableError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class YubiKeyNotPressedTimeout(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)


class ConfigMigrationError(AtbuException):
    def __init__(self, message: str = None, cause=None):
        self._cause = cause
        super().__init__(message=message, cause=cause)



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import setuptools

# Redirect from long_description to github README.rst since
# PyPi chokes on this particular README.rst.
# with open("README.rst", "r", encoding="utf-8") as fh:
#    long_description = fh.read()

setuptools.setup(
    name="atbu-pkg",
    version="0.0.25",
    author="Ashley R. Thomas",
    author_email="ashley.r.thomas.701@gmail.com",
    description= (
        "ATBU package supports local/cloud backup/restore "
        "as well as local file integrity diff tool for helping in "
        "efforts to manage file integrity, duplication, and bitrot detection."
    ),
    entry_points = {
        'console_scripts': ['atbu=atbu.tools.backup.command_line:main']
    },
    long_description="""
ATBU Backup & Persistent File Information is a local/cloud backup/restore
command-line utility with optional deduplication and bitrot detection,
plus a little utility with useful digest-based directory file diff'ing.

Install: `pip install atbu-pkg`

GitHub: https://github.com/AshleyT3/atbu

Documentation: https://atbu.readthedocs.io/en/latest/
""",
    long_description_content_type="text/markdown",
    url="https://github.com/AshleyT3/atbu",
    project_urls={
        "Bug Tracker": "https://github.com/AshleyT3/atbu/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_namespace_packages(where="src"),
    python_requires=">=3.9",
    install_requires=[
        "atbu-common-pkg >= 0.0.6",
        "atbu-mp-pipeline-pkg >= 0.0.9",
        "cryptography >= 36.0.2",
        "pwinput >= 1.0.2",
        "keyring >= 23.5.0",
        "apache-libcloud >= 3.5.1",
        "azure-storage-blob >= 12.16.0",
        "google-auth >= 2.6.5",
        "google-cloud-storage >= 2.3.0",
        "google-resumable-media >= 2.3.2",
        "tabulate >= 0.8.9",
        "Send2Trash >= 1.8.0",
    ]
)



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Cloud storage basic E2E.
"""

# pylint: disable=unused-argument
# pylint: disable=unused-variable
# pylint: disable=unused-import

import os
from pathlib import Path
import logging
import keyring
import pytest

# Importing google.api_core.exceptions acts as a
# workaround for an issue where, if it is used
# by a factory used by a test, and loaded dynamically
# the first test will succeed, the others failing.
# This only happens with pytest and seems to relate
# to stale state in protobuf message.cc in relation to
# how the environment is managed when pytester is used,
# where id(_message.Message) changes at Python level, but
# not in the message.cc related pyd (dll).
import google.api_core.exceptions

LOGGER = logging.getLogger(__name__)

# pylint: disable=unused-import,wrong-import-position
from pytest import (
    LogCaptureFixture,
    CaptureFixture,
    fail,
    raises,
    Pytester,
    FixtureRequest,
    Config,
    RunResult,
    ExitCode,
)

from atbu.tools.backup.constants import (
    CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD,
    CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION,
    CONFIG_PASSWORD_KIND_FILENAME,
    CONFIG_PASSWORD_KIND_ACTUAL,
    CONFIG_VALUE_NAME_CONTAINER,
    CREDS_SUBCMD_CREATE_STORAGE_DEF,
)
from atbu.tools.backup.config import AtbuConfig

from atbu.tools.backup.storage_interface.base import StorageInterfaceFactory

# pylint: disable=unused-import
from .secrets import (
    AWS_STORAGE_ACCESSKEY,
    AWS_STORAGE_SECRET,
    AZSDK_AZURE_BLOB_STORAGE_SECRET,
    AZSDK_AZURE_BLOB_STORAGE_USERKEY,
    LIBCLOUD_AZURE_BLOB_STORAGE_SECRET,
    LIBCLOUD_AZURE_BLOB_STORAGE_USERKEY,
    GOOGLE_STORAGE_ACCESS_KEY,
    GOOGLE_STORAGE_SECRET,
    GOOGLE_STORAGE_SERVICE_ACCOUNT_CLIENT_EMAIL,
    GOOGLE_STORAGE_SERVICE_ACCOUNT_JSON_PATH,
    GOOGLE_STORAGE_SERVICE_ACCOUNT_PROJECT_ID,
)

from .common_helpers import (
    create_test_data_directory_minimal_vary,
    establish_random_seed,
    create_test_data_directory_basic,
    create_test_data_directory_minimal,
    DirInfo,
    run_atbu,
    directories_match_entirely_by_order,
    extract_storage_definition_and_config_file_path,
    validate_backup_recovery,
    validate_backup_restore,
    validate_backup_restore_history,
    validate_cred_export_import,
)

TEST_BACKUP_NAME = "atbu-backup-5b497bb3-c9ef-48a9-af7b-2327fc17fb65"
TEST_CONTAINER_BASE_NAME = "atbu-bucket"

# import pdb; pdb.set_trace()
# import pdb; pdb.set_trace()


def setup_module(module):  # pylint: disable=unused-argument
    pass


def teardown_module(module):  # pylint: disable=unused-argument
    pass


def get_storage_factory(storage_def_name: str) -> StorageInterfaceFactory:
    return StorageInterfaceFactory.create_factory_from_storage_def_name(
        storage_def_name=storage_def_name
    )


def get_storage_def_dict(storage_def_name: str) -> dict:
    cfg: AtbuConfig
    cfg, _, storage_def_dict = AtbuConfig.access_cloud_storage_config(
        storage_def_name=storage_def_name,
        must_exist=True,
        create_if_not_exist=False,
    )
    return storage_def_dict


def get_all_storage_def_names() -> list[str]:
    return AtbuConfig.get_user_storage_def_names()


def get_container_name(storage_def_name: str) -> str:
    storage_def_dict = get_storage_def_dict(storage_def_name=storage_def_name)
    return storage_def_dict[CONFIG_VALUE_NAME_CONTAINER]


def create_storage_definition_json(
    interface,
    provider,
    project_id,
    userkey,
    secret,
    tmp_path: Path,
    pytester: Pytester,
    use_secrets_prompt: bool = False,
) -> tuple[str, str, str]:

    if use_secrets_prompt:
        # No secrets on command line.
        driver_arg = ""
        # Input secrets...
        # access key<ENTER>
        # secret access key<ENTER>
        # ENTER ENTER to accept both defaults for backup encryption.
        stdin_resp_enable_enc_pwd_not_req = (
            f"{userkey}{os.linesep}" +
            f"{secret}{os.linesep}" +
            f"{os.linesep}{os.linesep}"
        ).encode()
    else:
        # Specify all driver arguments, including secrets, on command line.
        driver_arg = f"key={userkey},secret={secret}"
        # ENTER ENTER to accept both defaults.
        stdin_resp_enable_enc_pwd_not_req = f"{os.linesep}{os.linesep}".encode()

    # Specify other cloud storage driver kv pairs on command line.
    # If project id is specified, add it to command line.
    if project_id is not None:
        driver_arg = (driver_arg + f",project={project_id}").strip(", ")

    if len(driver_arg) > 0:
        # Specify driver_arg.
        rr = run_atbu(
            pytester,
            tmp_path,
            "creds",
            CREDS_SUBCMD_CREATE_STORAGE_DEF,
            TEST_BACKUP_NAME,
            interface,
            provider,
            f"{TEST_CONTAINER_BASE_NAME}*",
            driver_arg,
            stdin=stdin_resp_enable_enc_pwd_not_req,
            log_base_name=f"{CREDS_SUBCMD_CREATE_STORAGE_DEF}-json",
        )
    else:
        # Do not specify driver_arg.
        rr = run_atbu(
            pytester,
            tmp_path,
            "creds",
            CREDS_SUBCMD_CREATE_STORAGE_DEF,
            TEST_BACKUP_NAME,
            interface,
            provider,
            f"{TEST_CONTAINER_BASE_NAME}*",
            stdin=stdin_resp_enable_enc_pwd_not_req,
            log_base_name=f"{CREDS_SUBCMD_CREATE_STORAGE_DEF}-json",
        )
    assert rr.ret == ExitCode.OK

    storage_def_name_atbu_cfg_path_list = (
        extract_storage_definition_and_config_file_path(rr.outlines)
    )
    assert len(storage_def_name_atbu_cfg_path_list) >= 1
    last_save = storage_def_name_atbu_cfg_path_list[-1]
    storage_def_name = last_save[0]
    atbu_cfg_path = last_save[1]

    cfg, _, _ = AtbuConfig.access_cloud_storage_config(
        storage_def_name=storage_def_name,
        must_exist=True,
        create_if_not_exist=False,
    )

    container_name = cfg.get_storage_def_dict(storage_def_name=storage_def_name)[
        CONFIG_VALUE_NAME_CONTAINER
    ]

    return storage_def_name, atbu_cfg_path, container_name


def delete_all_objects(storage_def_name: str):
    factory = get_storage_factory(storage_def_name=storage_def_name)

    interface = factory.create_storage_interface()
    container_name = get_container_name(storage_def_name=storage_def_name)
    container = interface.get_container(container_name=container_name)
    list_objs = container.list_objects()
    for obj in list_objs:
        container.delete_object(object_name=obj.name)


def delete_container(storage_def_name: str):
    factory = get_storage_factory(storage_def_name=storage_def_name)
    interface = factory.create_storage_interface()
    container_name = get_container_name(storage_def_name=storage_def_name)

    delete_all_objects(storage_def_name=storage_def_name)

    interface.delete_container(container_name=container_name)


def delete_all_containers():
    all_storage_def_names = get_all_storage_def_names()
    for name in all_storage_def_names:
        try:
            delete_container(storage_def_name=name)
        except Exception as ex:
            LOGGER.error(f"failed to delete container {name}. {ex}")


def delete_storage_definition_json(
    tmp_path: Path,
    pytester: Pytester,
):
    delete_container(storage_def_name=TEST_BACKUP_NAME)

    # Enter Y for yes to confirm storage def deletion.
    # Enter Y for yes to confirm deletion of backup information.
    stdin_resp_enter_y_for_yes = f"y{os.linesep}y{os.linesep}".encode()

    rr = run_atbu(
        pytester,
        tmp_path,
        "creds",
        "delete-storage-def",
        TEST_BACKUP_NAME,
        stdin=stdin_resp_enter_y_for_yes,
        log_base_name="delete-storage-def",
    )
    assert rr.ret == ExitCode.OK

    kr_pwd_after_del = keyring.get_password(
        service_name=TEST_BACKUP_NAME, username=CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD
    )
    assert kr_pwd_after_del is None

    kr_pwd_after_del = keyring.get_password(
        service_name=TEST_BACKUP_NAME,
        username=CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION,
    )
    assert kr_pwd_after_del is None


@pytest.fixture(autouse=True)
def cleanup_keyring(pytester: Pytester):
    yield

    try:
        delete_all_containers()
    except Exception:
        pass

    try:
        keyring.delete_password(
            service_name=TEST_BACKUP_NAME,
            username=CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD,
        )
    except Exception:
        pass

    try:
        keyring.delete_password(
            service_name=TEST_BACKUP_NAME,
            username=CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION,
        )
    except Exception:
        pass


backup_restore_parameters = [
    pytest.param(
        "google",
        "google_storage",
        GOOGLE_STORAGE_SERVICE_ACCOUNT_PROJECT_ID,
        GOOGLE_STORAGE_SERVICE_ACCOUNT_CLIENT_EMAIL,
        GOOGLE_STORAGE_SERVICE_ACCOUNT_JSON_PATH,
        CONFIG_PASSWORD_KIND_FILENAME,
        id="google",
        marks=pytest.mark.skipif(
            GOOGLE_STORAGE_SERVICE_ACCOUNT_CLIENT_EMAIL == "skip",
            reason="secrets not available.",
        ),
    ),
    pytest.param(
        "azure",
        "azure_blobs",
        None,
        AZSDK_AZURE_BLOB_STORAGE_USERKEY,
        AZSDK_AZURE_BLOB_STORAGE_SECRET,
        CONFIG_PASSWORD_KIND_ACTUAL,
        id="azsdk_azure",
        marks=pytest.mark.skipif(
            AZSDK_AZURE_BLOB_STORAGE_USERKEY == "skip", reason="secrets not available."
        ),
    ),
    pytest.param(
        "libcloud",
        "azure_blobs",
        None,
        LIBCLOUD_AZURE_BLOB_STORAGE_USERKEY,
        LIBCLOUD_AZURE_BLOB_STORAGE_SECRET,
        CONFIG_PASSWORD_KIND_ACTUAL,
        id="libcloud_azure",
        marks=pytest.mark.skipif(
            LIBCLOUD_AZURE_BLOB_STORAGE_USERKEY == "skip", reason="secrets not available."
        ),
    ),
    pytest.param(
        "libcloud",
        "s3",
        None,
        AWS_STORAGE_ACCESSKEY,
        AWS_STORAGE_SECRET,
        CONFIG_PASSWORD_KIND_ACTUAL,
        id="aws",
        marks=pytest.mark.skipif(
            AWS_STORAGE_ACCESSKEY == "skip", reason="secrets not available."
        ),
    ),
]


@pytest.mark.parametrize(
    "interface,provider,project_id,userkey,secret,secret_type",
    backup_restore_parameters,
)
def test_create_storage_definition_json(
    interface,
    provider,
    project_id,
    userkey,
    secret,
    secret_type,
    tmp_path: Path,
    pytester: Pytester,
):
    storage_def_name, atbu_cfg_path, container_name = create_storage_definition_json(
        interface,
        provider,
        project_id,
        userkey,
        secret,
        tmp_path=tmp_path,
        pytester=pytester,
    )
    assert storage_def_name == TEST_BACKUP_NAME
    assert os.path.isfile(atbu_cfg_path)

    cfg, _, _ = AtbuConfig.access_cloud_storage_config(
        storage_def_name=storage_def_name,
        must_exist=True,
        create_if_not_exist=False,
    )
    storage_def_dict = cfg.get_storage_def_with_resolved_secrets_deep_copy(
        storage_def_name=TEST_BACKUP_NAME, keep_secrets_base64_encoded=False
    )
    assert container_name.startswith(TEST_CONTAINER_BASE_NAME)
    assert len(container_name) > len(TEST_CONTAINER_BASE_NAME)
    client_email_from_cfg = storage_def_dict["driver"]["key"]
    if project_id is not None:
        project_id_from_cfg = storage_def_dict["driver"]["project"]
        assert project_id_from_cfg == project_id
    secret_from_cfg = storage_def_dict["driver"]["secret"].decode("utf-8")
    password_type_from_cfg = storage_def_dict["driver"]["password_type"]
    assert client_email_from_cfg == userkey
    assert secret_from_cfg == secret
    assert password_type_from_cfg == secret_type

    delete_storage_definition_json(tmp_path=tmp_path, pytester=pytester)
    pass  # pylint: disable=unnecessary-pass

@pytest.mark.parametrize(
    "interface,provider,project_id,userkey,secret,secret_type",
    backup_restore_parameters,
)
def test_create_storage_definition_json__secrets_prompt(
    interface,
    provider,
    project_id,
    userkey,
    secret,
    secret_type,
    tmp_path: Path,
    pytester: Pytester,
):
    storage_def_name, atbu_cfg_path, container_name = create_storage_definition_json(
        interface,
        provider,
        project_id,
        userkey,
        secret,
        tmp_path=tmp_path,
        pytester=pytester,
        use_secrets_prompt=True,
    )
    assert storage_def_name == TEST_BACKUP_NAME
    assert os.path.isfile(atbu_cfg_path)

    cfg, _, _ = AtbuConfig.access_cloud_storage_config(
        storage_def_name=storage_def_name,
        must_exist=True,
        create_if_not_exist=False,
    )
    storage_def_dict = cfg.get_storage_def_with_resolved_secrets_deep_copy(
        storage_def_name=TEST_BACKUP_NAME, keep_secrets_base64_encoded=False
    )
    assert container_name.startswith(TEST_CONTAINER_BASE_NAME)
    assert len(container_name) > len(TEST_CONTAINER_BASE_NAME)
    client_email_from_cfg = storage_def_dict["driver"]["key"]
    if project_id is not None:
        project_id_from_cfg = storage_def_dict["driver"]["project"]
        assert project_id_from_cfg == project_id
    secret_from_cfg = storage_def_dict["driver"]["secret"].decode("utf-8")
    password_type_from_cfg = storage_def_dict["driver"]["password_type"]
    assert client_email_from_cfg == userkey
    assert secret_from_cfg == secret
    assert password_type_from_cfg == secret_type

    delete_storage_definition_json(tmp_path=tmp_path, pytester=pytester)
    pass  # pylint: disable=unnecessary-pass


@pytest.mark.parametrize(
    "interface,provider,project_id,userkey,secret,secret_type",
    backup_restore_parameters,
)
def test_backup_restore(
    interface,
    provider,
    project_id,
    userkey,
    secret,
    secret_type,
    tmp_path: Path,
    pytester: Pytester,
):
    storage_def_name, atbu_cfg_path, container_name = create_storage_definition_json(
        interface,
        provider,
        project_id,
        userkey,
        secret,
        tmp_path=tmp_path,
        pytester=pytester,
    )
    assert storage_def_name == TEST_BACKUP_NAME
    assert os.path.isfile(atbu_cfg_path)

    establish_random_seed(tmp_path)

    source_directory = tmp_path / "SourceDataDir"

    _, files_created = create_test_data_directory_basic(
        path_to_dir=source_directory,
    )
    total_files = len(files_created)
    assert total_files > 0

    storage_specifier = f"storage:{TEST_BACKUP_NAME}"

    validate_backup_restore(
        pytester=pytester,
        tmp_path=tmp_path,
        source_directory=source_directory,
        expected_total_files=total_files,
        storage_specifier=storage_specifier,
        compression_type="normal",
        backup_timeout=60 * 5,
        restore_timeout=60 * 5,
        initial_backup_stdin=None,
    )

    delete_storage_definition_json(tmp_path=tmp_path, pytester=pytester)
    pass  # pylint: disable=unnecessary-pass


@pytest.mark.parametrize(
    "interface,provider,project_id,userkey,secret,secret_type",
    backup_restore_parameters,
)
def test_backup_restore__secrets_prompt(
    interface,
    provider,
    project_id,
    userkey,
    secret,
    secret_type,
    tmp_path: Path,
    pytester: Pytester,
):
    storage_def_name, atbu_cfg_path, container_name = create_storage_definition_json(
        interface,
        provider,
        project_id,
        userkey,
        secret,
        tmp_path=tmp_path,
        pytester=pytester,
        use_secrets_prompt=True,
    )
    assert storage_def_name == TEST_BACKUP_NAME
    assert os.path.isfile(atbu_cfg_path)

    establish_random_seed(tmp_path)

    source_directory = tmp_path / "SourceDataDir"

    _, files_created = create_test_data_directory_minimal_vary(
        path_to_dir=source_directory,
    )
    total_files = len(files_created)
    assert total_files > 0

    storage_specifier = f"storage:{TEST_BACKUP_NAME}"

    validate_backup_restore(
        pytester=pytester,
        tmp_path=tmp_path,
        source_directory=source_directory,
        expected_total_files=total_files,
        storage_specifier=storage_specifier,
        compression_type="normal",
        backup_timeout=60 * 5,
        restore_timeout=60 * 5,
        initial_backup_stdin=None,
    )

    delete_storage_definition_json(tmp_path=tmp_path, pytester=pytester)
    pass  # pylint: disable=unnecessary-pass


@pytest.mark.parametrize(
    "interface,provider,project_id,userkey,secret,secret_type",
    backup_restore_parameters,
)
def test_backup_restore_history(
    interface,
    provider,
    project_id,
    userkey,
    secret,
    secret_type,
    tmp_path: Path,
    pytester: Pytester,
):
    storage_def_name, atbu_cfg_path, container_name = create_storage_definition_json(
        interface,
        provider,
        project_id,
        userkey,
        secret,
        tmp_path=tmp_path,
        pytester=pytester,
    )
    assert storage_def_name == TEST_BACKUP_NAME
    assert os.path.isfile(atbu_cfg_path)

    establish_random_seed(tmp_path)

    source_directory = tmp_path / "SourceDataDir"

    _, files_created = create_test_data_directory_minimal_vary(
        path_to_dir=source_directory,
    )
    total_files = len(files_created)
    assert total_files > 0

    storage_specifier = f"storage:{TEST_BACKUP_NAME}"

    validate_backup_restore_history(
        pytester=pytester,
        tmp_path=tmp_path,
        max_history=3,
        source_directory=source_directory,
        expected_total_files=total_files,
        storage_specifier=storage_specifier,
        compression_type="normal",
        backup_timeout=60 * 5,
        restore_timeout=60 * 5,
        initial_backup_stdin=None,
    )

    delete_storage_definition_json(tmp_path=tmp_path, pytester=pytester)
    pass  # pylint: disable=unnecessary-pass


@pytest.mark.parametrize(
    "interface,provider,project_id,userkey,secret,secret_type",
    backup_restore_parameters,
)
def test_credential_export_import(
    interface,
    provider,
    project_id,
    userkey,
    secret,
    secret_type,
    tmp_path: Path,
    pytester: Pytester,
):
    storage_def_name, atbu_cfg_path, container_name = create_storage_definition_json(
        interface,
        provider,
        project_id,
        userkey,
        secret,
        tmp_path=tmp_path,
        pytester=pytester,
    )
    assert storage_def_name == TEST_BACKUP_NAME
    assert os.path.isfile(atbu_cfg_path)

    establish_random_seed(tmp_path)

    source_directory = tmp_path / "SourceDataDir"

    _, files_created = create_test_data_directory_minimal(
        path_to_dir=source_directory,
    )
    total_files = len(files_created)
    assert total_files > 0

    #
    # Backup file. Restoring this, or not, is basis for other
    # results below.
    #
    rr = run_atbu(
        pytester,
        tmp_path,
        "backup",
        "--full",
        source_directory,
        f"storage:{storage_def_name}",
        timeout=60 * 5,
        log_base_name="backup",
    )
    assert rr.ret == ExitCode.OK

    validate_cred_export_import(
        pytester=pytester,
        tmp_path=tmp_path,
        atbu_cfg_path=atbu_cfg_path,
        source_directory=source_directory,
        expected_total_files=total_files,
        storage_def_name=storage_def_name,
        storage_specifier=f"storage:{storage_def_name}",
    )

    delete_storage_definition_json(tmp_path=tmp_path, pytester=pytester)
    pass  # pylint: disable=unnecessary-pass


@pytest.mark.parametrize(
    "interface,provider,project_id,userkey,secret,secret_type",
    backup_restore_parameters,
)
def test_credential_and_backup_info_recovery(
    interface,
    provider,
    project_id,
    userkey,
    secret,
    secret_type,
    tmp_path: Path,
    pytester: Pytester,
):
    storage_def_name, atbu_cfg_path, container_name = create_storage_definition_json(
        interface,
        provider,
        project_id,
        userkey,
        secret,
        tmp_path=tmp_path,
        pytester=pytester,
    )
    assert storage_def_name == TEST_BACKUP_NAME
    assert os.path.isfile(atbu_cfg_path)

    establish_random_seed(tmp_path)

    source_directory = tmp_path / "SourceDataDir"

    _, files_created = create_test_data_directory_minimal(
        path_to_dir=source_directory,
    )
    total_files = len(files_created)
    assert total_files > 0

    #
    # Backup file. Restoring this, or not, is basis for other
    # results below.
    #
    rr = run_atbu(
        pytester,
        tmp_path,
        "backup",
        "--full",
        source_directory,
        f"storage:{storage_def_name}",
        timeout=60 * 5,
        log_base_name="backup",
    )
    assert rr.ret == ExitCode.OK

    validate_backup_recovery(
        pytester=pytester,
        tmp_path=tmp_path,
        atbu_cfg_path=atbu_cfg_path,
        source_directory=source_directory,
        expected_total_files=total_files,
        storage_def_name=storage_def_name,
        storage_specifier=f"storage:{storage_def_name}",
    )

    delete_storage_definition_json(tmp_path=tmp_path, pytester=pytester)
    pass  # pylint: disable=unnecessary-pass



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Singleton global hasher for the process.
"""

from atbu.common.exception import SingletonAlreadyCreated
from atbu.common.hasher import (
    DEFAULT_HASH_ALGORITHM,
    HasherDefinitions,
)
from atbu.common.singleton import Singleton


class GlobalHasherDefinitions(HasherDefinitions, metaclass=Singleton):
    _init_algos: list[str] = None

    def __init__(self, algos: list[str] = None):
        if self._init_algos:
            if algos:
                raise SingletonAlreadyCreated(
                    "Access to singleton after initial creation must have no arguments."
                )
            return  # do not call base class __init__ after first-time init.
        if not algos:
            algos = [DEFAULT_HASH_ALGORITHM]
        if not isinstance(algos, list):
            raise ValueError(
                f"GlobalHasherDefinitions: The algos argument must be a list of algorithms."
            )
        self._init_algos = algos
        super().__init__(algos)



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Classes/functions related to backup-related selections.
"""

from dataclasses import dataclass, field
from fnmatch import fnmatchcase, fnmatch
import logging
import os
from pathlib import Path
import platform
import re

from atbu.common.exception import (
    InvalidStateError,
)
from atbu.common.util_helpers import (
    iwalk_fnmatch,
)
from .exception import *
from .constants import *
from .config import (
    AtbuConfig,
    is_existing_filesystem_storage_path,
)
from .backup_core import (
    StorageDefinition,
    SpecificBackupInformation,
    BackupInformationDatabase,
)
from .backup_core import BackupFileInformation

DEFAULT_PLATFORM_EXCLUDE = {
    "Windows": {
        re.compile(os.path.normcase(r"^[a-zA-Z]:\\System Volume Information" "\\\\")),
        re.compile(os.path.normcase(r"^[a-zA-Z]:\\$Recycle.Bin" "\\\\")),
    }
}

SELECTION_SCOPE_BACKUP = "backup"
SELECTION_SCOPE_FILE = "files"
SELECTION_SCOPE_ALL = "all"
SELECTION_SCOPES = [SELECTION_SCOPE_ALL, SELECTION_SCOPE_BACKUP, SELECTION_SCOPE_FILE]

SELECTION_BACKUP_LAST = "last"


class SelectionPattern:
    def __init__(self, raw_command_line_specifier: str):
        parts = raw_command_line_specifier.split(":", maxsplit=1)
        scope = None
        pattern = parts
        if len(parts) > 1:
            scope = parts[0]
            pattern = parts[1]
        if scope not in SELECTION_SCOPES:
            raise BackupSelectionError(
                f"The scope '{scope}' is invalid, must be one of {SELECTION_SCOPES}."
            )
        self.scope = scope
        self.original_pattern = pattern
        self.normcase_pattern = os.path.normcase(pattern)

    @staticmethod
    def is_valid_selection_pattern(arg_to_test: str):
        parts = arg_to_test.split(":", maxsplit=1)
        return len(parts) == 2 and parts[0] in SELECTION_SCOPES


@dataclass
class StorageSelectionInfo:
    location: str
    location_derivation_source_arg: str
    storage_def_name: str
    storage_def: dict
    backup_info_dir: Path
    selection_patterns: list[SelectionPattern]


@dataclass
class SpecificBackupSelection:
    storage_def_name: str
    specific_backup_name: str
    specific_backup_info: SpecificBackupInformation
    storage_def: StorageDefinition
    backup_info_dir: Path
    backup_history: BackupInformationDatabase
    sel_info: StorageSelectionInfo
    selected_fi: dict[str, BackupFileInformation] = field(default_factory=dict)


def parse_storage_def_specifiers_patterns(
    raw_arguments: list[str], resolve_storage_def_secrets: bool = False
) -> list[StorageSelectionInfo]:
    sel_info_dict: dict[str, StorageSelectionInfo] = {}
    result: list[StorageSelectionInfo] = []
    cur_storage_selections: list[StorageSelectionInfo] = None
    for raw_arg in raw_arguments:
        if SelectionPattern.is_valid_selection_pattern(arg_to_test=raw_arg):
            #
            # Valid selection pattern, apply it to cur_storage_selections.
            #
            if not cur_storage_selections:
                raise StorageDefinitionNotSpecifiedError(
                    f"No storage selection to apply pattern '{raw_arg}' to."
                )
            for (
                cur_storage_selection
            ) in cur_storage_selections:  # pylint: disable=not-an-iterable
                cur_storage_selection.selection_patterns.append(
                    SelectionPattern(raw_command_line_specifier=raw_arg)
                )
            continue

        storage_pattern = raw_arg.lower()
        raw_arg_split = raw_arg.split(":", maxsplit=1)
        if (
            raw_arg_split[0]
            == CONFIG_SECTION_STORAGE_DEFINITION_SPECIFIER_PREFIX[
                : len(raw_arg_split[0])
            ]
        ):
            storage_pattern = raw_arg_split[1]

        storage_specifiers = [
            f"{CONFIG_SECTION_STORAGE_DEFINITION_SPECIFIER_PREFIX}:{n}"
            for n in AtbuConfig.get_user_storage_def_names(fnmatch_pattern=storage_pattern)
        ]

        if not storage_specifiers:
            if not is_existing_filesystem_storage_path(storage_location=raw_arg):
                raise StorageDefinitionNotFoundError(
                    f"The specifier '{raw_arg}' did not "
                    f"yield any storage definitions or "
                    f"filesystem backups."
                )
            # The raw_arg is a directory, set it to the specifier to allow following
            # to try to find a configuration.
            storage_specifiers = [raw_arg]

        #
        # Fall through, raw_arg is not a selection pattern,
        # should be storage definition specifier...
        #

        # Per-iteration storage selections.
        cur_storage_selections: list[StorageSelectionInfo] = []

        # For each specifier, find a storage, add to cur_storage_selections
        # for extracting any patterns.
        for storage_specifier in storage_specifiers:
            atbu_cfg_to_use: AtbuConfig
            (
                atbu_cfg_to_use,
                storage_def_name,
                storage_def_dict,
            ) = AtbuConfig.resolve_storage_location(
                storage_location=storage_specifier,
                resolve_storage_def_secrets=resolve_storage_def_secrets,
                create_if_not_exist=False,
            )
            if not atbu_cfg_to_use or not storage_def_name or not storage_def_dict:
                raise StorageDefinitionNotFoundError(
                    f"WARNING: Storage definition for '{storage_specifier}' not found, skipping."
                )

            storage_sel_info: StorageSelectionInfo = sel_info_dict.get(storage_def_name)
            if not storage_sel_info:
                storage_def = StorageDefinition.storage_def_from_dict(
                    storage_def_name=storage_def_name, storage_def_dict=storage_def_dict
                )

                storage_sel_info = StorageSelectionInfo(
                    location=storage_specifier,
                    location_derivation_source_arg=raw_arg,
                    storage_def_name=storage_def_name,
                    storage_def=storage_def,
                    backup_info_dir=atbu_cfg_to_use.get_primary_backup_info_dir(),
                    selection_patterns=[],
                )
                # Add to ordered list of StorageSelectionInfo instance results.
                result.append(storage_sel_info)

            # Add to the StorageSelectionInfo instances related to
            # the upcoming SelectionPattern instances (further above).
            # A single StorageSelectionInfo instance could be added to
            # cur_storage_selections twice if perhaps the same storage def
            # was specified twice in the overall options.
            cur_storage_selections.append(storage_sel_info)

    # Resulting list should only have any given storage defintion once,
    # with any patterns extracted per above.
    return result


def get_selections_from_file_info_dict(
    sel_pat: SelectionPattern, nc_path_to_fi_dict: dict[str, BackupFileInformation]
) -> list[BackupFileInformation]:
    r: list[BackupFileInformation] = []
    # Check each normcase'ed path against the glob-like pattern.
    for normcase_path, fi in nc_path_to_fi_dict.items():
        # path and pattern are already normcase so using fnmatchcase is fine.
        if not fnmatchcase(normcase_path, sel_pat.normcase_pattern):
            continue
        # Sanity check state, ensure file is resolved with backing_fi.
        if fi.is_unchanged_since_last and not fi.backing_fi:
            # If unchanged file has unresolved backing_fi, fail.
            raise BackingFileInformationNotFound(
                f"An unchanged since last file has no backing file information: "
                f"fi={fi.path}"
            )
        r.append(fi)
    return r


def get_specific_backup_selections(
    sel_info_list: list[StorageSelectionInfo],
) -> list[list[SpecificBackupSelection]]:
    result: list[list[SpecificBackupSelection]] = []
    history_dict: dict[str, BackupInformationDatabase] = {}
    # For each storage definition being filtered...
    for si in sel_info_list:
        if not si.selection_patterns:
            # No filters, skip.
            # TODO: Should this default to none or * all ? (TBD)
            # For now, create a dummy SpecificBackupSelection to allow
            # querying of storage defintions via this code path.
            specific_backup_sel = SpecificBackupSelection(
                storage_def_name=si.storage_def_name,
                specific_backup_name=None,
                specific_backup_info=None,
                storage_def=si.storage_def,
                backup_info_dir=si.backup_info_dir,
                backup_history=None,
                sel_info=si,
            )
            result.append([specific_backup_sel])
            continue
        # For the si storage definition, extract all selected specific backup instances.
        specific_backup_selections: list[SpecificBackupSelection] = []
        for sp in si.selection_patterns:
            if sp.scope not in [SELECTION_SCOPE_BACKUP]:
                continue
            #
            # Share the same storage definition history with SpecificBackupSelection
            # instances of the same storage definition as follows.
            #
            history_key = f"{si.storage_def_name}:{si.backup_info_dir}"
            backup_history = history_dict.get(history_key)
            if not backup_history:
                backup_history = BackupInformationDatabase.create_from_file(
                    backup_base_name=si.storage_def_name,
                    backup_info_dir=si.backup_info_dir,
                )
            # For each specific backup, from newest to oldest...
            for sbi in backup_history.get_specific_backups():
                # If user specified SELECTION_BACKUP_LAST, always match and
                # select the first most recent specific backup.
                # If the user specified something actual fnmatch criteria and
                # it matches, select the backup.
                # All other cases skip this specific backup.
                if (
                    not sp.original_pattern.lower() == SELECTION_BACKUP_LAST
                    and not fnmatch(sbi.specific_backup_name, sp.normcase_pattern)
                ):
                    continue
                # Match, backup selections.
                specific_backup_sel = SpecificBackupSelection(
                    storage_def_name=si.storage_def_name,
                    specific_backup_name=sbi.specific_backup_name,
                    specific_backup_info=sbi,
                    storage_def=si.storage_def,
                    backup_info_dir=si.backup_info_dir,
                    backup_history=backup_history,
                    sel_info=si,
                )
                specific_backup_selections.append(specific_backup_sel)
                if sp.original_pattern.lower() == SELECTION_BACKUP_LAST:
                    # With selector SELECTION_BACKUP_LAST skip backups
                    # older than the last one (the most recent one).
                    break
        for sp in si.selection_patterns:
            if sp.scope != SELECTION_SCOPE_FILE:
                continue
            specific_backup_sel: SpecificBackupSelection
            # For each selected specific backup, from newest to oldest...
            for specific_backup_sel in specific_backup_selections:
                # Place all specific backup file info into a dict keyed w/normcase path.
                sb_nc_path_to_fi_dict: dict[str, BackupFileInformation] = {}
                for sb_fi in specific_backup_sel.specific_backup_info.all_file_info:
                    sb_nc_path_to_fi_dict[sb_fi.nc_path_without_root] = sb_fi
                # Determine which specific backup file info matches the file selection pattern.
                list_sel_fi = get_selections_from_file_info_dict(
                    sel_pat=sp, nc_path_to_fi_dict=sb_nc_path_to_fi_dict
                )
                # For each matching specific backup file info, if it has not already
                # matched the same path with a newer file info, add it to the selection.
                # There should not be duplicate paths per specific backup.
                # Callers will use functions like get_all_specific_backup_file_info to
                # reduce all selections across all specific backups to a single of duplicates
                # where applicable. That is not what the following does. The following is a
                # sanity check.
                for sel_fi in list_sel_fi:
                    if not specific_backup_sel.selected_fi.get(
                        sel_fi.nc_path_without_root
                    ):
                        specific_backup_sel.selected_fi[
                            sel_fi.nc_path_without_root
                        ] = sel_fi
        if specific_backup_selections:
            result.append(specific_backup_selections)
    return result


def verify_specific_backup_selection_list(sbs_list: list[SpecificBackupSelection]):
    storage_def_name = None
    backup_info_dir = None
    for sbs in sbs_list:
        if not storage_def_name:
            storage_def_name = sbs.storage_def_name
            backup_info_dir = sbs.backup_info_dir
        elif (
            storage_def_name != sbs.storage_def_name
            or backup_info_dir != sbs.backup_info_dir
        ):
            raise InvalidStateError(
                f"All SpecificBackupSelection instances in the list "
                f"must target the same storage definition."
            )


def get_all_specific_backup_file_info(
    sbs_list: list[SpecificBackupSelection],
) -> list[BackupFileInformation]:
    sb_nc_path_to_fi_dict: dict[str, BackupFileInformation] = {}
    # The sbs_list contains SpecificBackupSelection instances from
    # the most recent to oldest specific backup examined as part of
    # the selection process. For each SpecificBackupSelection instance
    # in that order...
    for sbs in sbs_list:
        # Go through the SpecificBackupSelection instance's items, if an item
        # is not in the result sb_nc_path_to_fi_dict, add it. This means
        # newest BackupFileInformation wins.
        for _, sbs_fi in sbs.selected_fi.items():
            if not sb_nc_path_to_fi_dict.get(sbs_fi.nc_path_without_root):
                sb_nc_path_to_fi_dict[sbs_fi.nc_path_without_root] = sbs_fi
    # Return a list of BackupFileInformation instances sorted by normcase path.
    # The sorting is not essential but during operations it can be nice to see
    # files processed with in an order making sense. Being nonessential, this
    # can be changed if needed.
    return sorted(
        [v for _, v in sb_nc_path_to_fi_dict.items()],
        key=lambda sb_fi: sb_fi.nc_path_without_root,
    )


def user_specifiers_to_selections(
    specifiers: list[str], no_selections_ok: bool = False
) -> list[list[SpecificBackupSelection]]:

    # Get selection information from all specifiers.
    sel_info_list = parse_storage_def_specifiers_patterns(
        raw_arguments=specifiers, resolve_storage_def_secrets=True
    )

    # Validate selection information.
    if not no_selections_ok:
        for sel_info in sel_info_list:
            if len(sel_info.selection_patterns) == 0:
                raise BackupSelectionError(
                    f"You must have at least one storage, backup, and file specifier. "
                    f"Example: storage:<storage_def_name> backup:last files:*"
                )

    # Get selections using selection information.
    sbs_list_list = get_specific_backup_selections(sel_info_list)

    if not no_selections_ok:
        # Validate selections.
        for sbs_list in sbs_list_list:
            file_info_count = 0
            for sbs in sbs_list:
                file_info_count += len(sbs.selected_fi)
            if file_info_count == 0:
                raise BackupSelectionError(
                    f"No selected files for {sbs_list[0].storage_def_name}. "
                    f"You must have at least one storage, backup, and file specifier. "
                    f"Example: storage:<storage_def_name> backup:last files:*"
                )

    # Everything seems OK, return selections to caller.
    return sbs_list_list


def get_storage_defintions_from_sbs_list_list(
    sbs_list_list: list[list[SpecificBackupSelection]],
) -> list[StorageDefinition]:
    storage_def_dict: dict[str, StorageDefinition] = {}
    for sbs_list in sbs_list_list:
        for sbs in sbs_list:
            storage_def_dict[sbs.storage_def_name] = sbs.storage_def
    storage_def_list: list[StorageDefinition] = sorted(
        storage_def_dict.values(), key=lambda sd: sd.storage_def_name
    )
    return storage_def_list


def is_system_excluded_path(path):
    platform_excludes = DEFAULT_PLATFORM_EXCLUDE.get(platform.system())
    if platform_excludes is not None:
        for pe in platform_excludes:
            if pe.match(os.path.normcase(path)):
                return True
    return False


RE_FIND_GLOB_CHARS = re.compile(r"[?\[\]*]")
PATH_SEP_CHARS = r"\/"
RE_CONTAINS_GLOB_CHARS = re.compile(".*[*?[].*")


def get_largest_path_without_wildcards(path: str):
    m = RE_FIND_GLOB_CHARS.search(path)
    if m:
        index = m.start()
        while path[index] not in PATH_SEP_CHARS:
            index -= 1
        if index >= 0:
            path = path[: index + 1]
    return path


def get_dirs_without_wildcards(locations: list[str]):
    results: list[str] = []
    for location in locations:
        location = get_largest_path_without_wildcards(location)
        results.append(os.path.abspath(location))
    return results


def ensure_glob_pattern_for_dir(dir_wc: str):
    """For a directory, if no glob pattern is present, ensure
    the glob "**" pattern is joined.
    """
    nc_dir_wc = os.path.normcase(dir_wc)
    nc_drive = os.path.splitdrive(nc_dir_wc)[0]
    if not RE_CONTAINS_GLOB_CHARS.match(dir_wc) and (
        os.path.isdir(dir_wc) or nc_dir_wc == nc_drive
    ):
        dir_wc = os.path.join(dir_wc, "**")
    return dir_wc


@dataclass
class SourceSpecifierDiscoveries:
    source_dir: str
    source_dir_wc: str
    discovered_paths: list[BackupFileInformation]


def get_local_file_information(
    src_dirs_wc: list[str], exclude_patterns: list[str]
) -> list[BackupFileInformation]:
    r"""Given locations containing one or more paths to search, return a list
    of discovered files.

    Each path in src_dirs_wc may or may not contain glob patterns. If a path in
    src_dirs_wc is a directory without a glob pattern, the '**' glob pattern
    will be appened. For example, "c:\abc" will become "c:\\abc\\**" to
    cause a recursive search of all files.

    This function will return a list of file information each of which will
    have its file_info.discovery_path set to the src_dirs_wc leading to its
    discovery.
    """

    exclude_patterns = list(map(os.path.normcase, exclude_patterns))

    def is_ignored(path):
        if is_system_excluded_path(path):
            logging.debug(f"Ignoring platform-excluded path: {path}")
            return True
        nonlocal exclude_patterns
        if len(exclude_patterns) == 0:
            return False
        path_nc = os.path.normcase(path)
        if any(map(lambda pat: fnmatchcase(name=path_nc, pat=pat), exclude_patterns)):
            logging.info(f"Ignoring user-excluded path: {path}")
            return True
        return False

    # Both src_dirs_wc and src_dirs_no_wc can be correlated by index.
    # Start by sorting caller's selections by length/normcase. Length
    # means discovery by shortest selector wins.
    src_dirs_wc.sort(key=lambda p: (len(p), os.path.normcase(p)))
    src_dirs_no_wc: list[str] = get_dirs_without_wildcards(locations=src_dirs_wc)

    # Ensure no duplicate paths in final selection.
    all_paths = set()

    # Track which src_dirs_wc selector discovered which file path.
    srcspec_to_disc: dict[str, SourceSpecifierDiscoveries] = {}

    # For each caller selection...
    for idx, src_dir_wc in enumerate(src_dirs_wc):

        # Get corresponding path without any glob patterns ("wildcards" or "wc").
        src_dir_no_wc = src_dirs_no_wc[idx]
        src_dir_no_wc_nc = os.path.normcase(src_dir_no_wc)

        # If location is either a directory or drive letter alone without
        # any glob patterns, then add glob pattern '**'.
        nc_location = os.path.normcase(src_dir_wc)
        nc_drive = os.path.splitdrive(nc_location)[0]
        if not RE_CONTAINS_GLOB_CHARS.match(src_dir_wc) and (
            os.path.isdir(src_dir_wc) or nc_location == nc_drive
        ):
            src_dir_wc = os.path.join(src_dir_wc, "**")

        # If user specified a single path/file, search dir is it's dir.
        src_dir_to_search = src_dir_no_wc
        if os.path.isfile(src_dir_no_wc) and src_dir_no_wc == src_dir_wc:
            src_dir_to_search, _ = os.path.split(src_dir_no_wc)
        # Search, removing anything already found.
        discovered = set(
            iwalk_fnmatch(
                root_no_wildcards=src_dir_to_search,
                fnmatch_pat=src_dir_wc,
            )
        ).difference(all_paths)

        # Refine further, retaining only files which are not ignored.
        discovered = [v for v in discovered if not is_ignored(v)]

        if len(discovered) == 0:
            # Nothing left, next item.
            continue

        # For the given search root folder (without glob pattern),
        # find an existing SourceSpecifierDiscoveries instance, if
        # non found, create one.
        src_spec_disc = srcspec_to_disc.get(src_dir_no_wc_nc)
        if not src_spec_disc:
            src_spec_disc = SourceSpecifierDiscoveries(
                source_dir=src_dir_no_wc, source_dir_wc=src_dir_wc, discovered_paths=[]
            )
            srcspec_to_disc[src_dir_no_wc_nc] = src_spec_disc

        # Add discovered files to instance, assocating them with
        # root path used to find the selections themselves.
        src_spec_disc.discovered_paths.extend(discovered)

        # Track files added, do not add more than once.
        all_paths.update(discovered)

    # Create results for caller by processing each SourceSpecifierDiscoveries instance.
    disc_list = sorted(
        srcspec_to_disc.values(), key=lambda d: os.path.normcase(d.source_dir)
    )
    file_info_list: list[BackupFileInformation] = []
    # For each caller-specified search path...
    for idx, disc in enumerate(disc_list):
        # For each file discovered by caller's search path...
        for p in disc.discovered_paths:
            if not os.path.normcase(p).startswith(os.path.normcase(disc.source_dir)):
                raise InvalidStateError(
                    f"Expected file to have path starting with discovery source: "
                    f"source_dir_wc={disc.source_dir_wc} "
                    f"file_info.path={p.path}"
                )
            # Create the result list with file info
            # indicating discovery source dir.
            file_info_list.append(
                BackupFileInformation(path=p, discovery_path=disc.source_dir)
            )
    return file_info_list
